"""This module manages the cache directory to run Docling MCP tools."""

import hashlib
import json
import os
import sys
from pathlib import Path
from typing import Optional

from docling_mcp.logger import setup_logger

# Create a default project logger
logger = setup_logger()


def hash_string(input_string: str) -> str:
    """Creates a hash-string from the input string."""
    return hashlib.sha256(input_string.encode(), usedforsecurity=False).hexdigest()


def get_cache_dir() -> Path:
    """Get the cache directory for the application.

    Returns:
        Path: A Path object pointing to the cache directory.

    The function will:
    1. First check for an environment variable 'CACHE_DIR'
    2. If not found, create a '_cache' directory in the root of the current package
    3. Ensure the directory exists before returning
    """
    # Check if cache directory is specified in environment variable
    cache_dir = os.environ.get("CACHE_DIR")

    if cache_dir:
        # Use the directory specified in the environment variable
        cache_path = Path(cache_dir)
    else:
        # Determine the package root directory
        if getattr(sys, "frozen", False):
            # Handle PyInstaller case
            package_root = Path(os.path.dirname(sys.executable))
        else:
            # Get the directory of the caller's module
            caller_file = sys._getframe(1).f_globals.get("__file__")

            if caller_file:
                # If running as a script or module
                current_path = Path(caller_file).resolve()

                # Find the package root by looking for the highest directory with an __init__.py
                package_root = current_path.parent
                while package_root.joinpath("__init__.py").exists():
                    package_root = package_root.parent
            else:
                # Fallback to current working directory if __file__ is not available
                package_root = Path.cwd()

        logger.info(f"package-root: {package_root}")

        # Create the cache directory path
        cache_path = package_root / "_cache"

    # Ensure cache directory exists
    logger.info(f"cache-path: {cache_path}")
    os.makedirs(cache_path, exist_ok=True)

    return cache_path


def get_cache_key(
    source: str, enable_ocr: bool = False, ocr_language: Optional[list[str]] = None
) -> str:
    """Generate a cache key for the document conversion."""
    key_data = {
        "source": source,
        "enable_ocr": enable_ocr,
        "ocr_language": ocr_language or [],
    }
    key_str = json.dumps(key_data, sort_keys=True)
    hash = hash_string(key_str)
    return hash[:32]



"""This module contains the settings for the Llama Stack usages."""

from pydantic_settings import BaseSettings, SettingsConfigDict


class Settings(BaseSettings):
    """Settings for the Llama Stack usages."""

    model_config = SettingsConfigDict(
        env_prefix="DOCLING_MCP_LLS_",
        env_file=".env",
        # extra="allow",
    )
    url: str = "http://localhost:8321"
    vdb_embedding: str = "all-MiniLM-L6-v2"
    extraction_model: str = "openai/gpt-oss-20b"


settings = Settings()



"""This module defines the RAG tools using Llama Stack."""

from dataclasses import dataclass
from typing import Annotated

from llama_stack_client.types.vector_io_insert_params import Chunk, ChunkChunkMetadata
from pydantic import Field
from transformers import AutoTokenizer

from docling_core.transforms.chunker.hierarchical_chunker import DocMeta
from docling_core.transforms.chunker.hybrid_chunker import HybridChunker
from docling_core.transforms.chunker.tokenizer.huggingface import HuggingFaceTokenizer

from docling_mcp.logger import setup_logger
from docling_mcp.settings.llama_stack import settings
from docling_mcp.shared import local_document_cache, mcp
from docling_mcp.tools.llama_stack._shared import get_llama_stack_client

logger = setup_logger()


@dataclass
class InsertDocumentOutput:
    """Output of the Docling document content insert tool."""

    vector_db_id: Annotated[
        str,
        Field(description="The vectordb identifier where the data has been inserted."),
    ]


@mcp.tool(title="Insert Docling document into a vector database")
def insert_document_to_vectordb(
    document_key: Annotated[
        str,
        Field(description="The unique identifier of the document in the local cache."),
    ],
    vector_db_id: Annotated[str, Field(description="The target vectordb identifier.")],
) -> InsertDocumentOutput:
    """Insert the document in the vectordb.

    This tool chunks and ingests a Docling document that exists in the local cache into
    a vectordb. The vectordb can then be used for knowledge searches.
    """
    logger.debug(f"{document_key=}")
    logger.debug(f"{vector_db_id=}")

    if document_key not in local_document_cache:
        doc_keys = ", ".join(local_document_cache.keys())
        raise ValueError(
            f"document-key: {document_key} is not found. Existing document-keys are: {doc_keys}"
        )

    doc = local_document_cache[document_key]
    doc_id = str(doc.origin.binary_hash) if doc.origin is not None else document_key

    tokenizer = HuggingFaceTokenizer(
        tokenizer=AutoTokenizer.from_pretrained(  # type: ignore
            pretrained_model_name_or_path=f"sentence-transformers/{settings.vdb_embedding}"
        ),
    )
    chunker = HybridChunker(tokenizer=tokenizer)

    chunk_iter = chunker.chunk(dl_doc=doc)

    ls_chunks: list[Chunk] = []
    for i, chunk in enumerate(chunk_iter):
        meta = DocMeta.model_validate(chunk.meta)

        chunk_id = f"{doc_id}-{i:05d}"

        enriched_text = chunker.contextualize(chunk=chunk)

        token_count = tokenizer.count_tokens(enriched_text)
        metadata = {
            "document_id": doc_id,
            "chunk_id": chunk_id,
            "token_count": token_count,
            # "metadata_token_count": 0,
            "doc_items": [item.self_ref for item in meta.doc_items],
        }
        chunk_metadata: ChunkChunkMetadata = {
            "document_id": doc_id,
            "chunk_id": chunk_id,
            "content_token_count": token_count,
        }
        chunk_dict: Chunk = {
            "content": enriched_text,
            "metadata": metadata,
            "chunk_metadata": chunk_metadata,
        }
        ls_chunks.append(chunk_dict)

    client = get_llama_stack_client()
    client.vector_io.insert(
        vector_db_id=vector_db_id,
        chunks=ls_chunks,
    )

    return InsertDocumentOutput(vector_db_id=vector_db_id)



"""Tools performing structured information extraction using Llama Stack model inference."""

import json
from dataclasses import dataclass
from typing import Annotated

from pydantic import Field

from docling_mcp.logger import setup_logger
from docling_mcp.settings.llama_stack import settings
from docling_mcp.shared import local_document_cache, mcp
from docling_mcp.tools.llama_stack._shared import get_llama_stack_client

logger = setup_logger()


@dataclass
class ExtractedContent:
    """Output of the requested key value pairs."""

    result: Annotated[
        dict[str, str],
        Field(
            description="Dictionary of the requested keys with the respective value extracted from the document."
        ),
    ]


@mcp.tool(title="Information extraction from a document")
def information_extraction(
    document_key: Annotated[
        str,
        Field(description="The unique identifier of the document in the local cache."),
    ],
    keys: Annotated[list[str], Field(description="List of keys to be extracted.")],
    descriptions: Annotated[
        list[str],
        Field(
            description="Description of the keys. Must be the same length of keys or it will be ignored."
        ),
    ],
    max_window_size: Annotated[
        int | None,
        Field(
            description="The maximum document size (in characters) to use for the extraction."
        ),
    ] = None,
) -> ExtractedContent:
    """Information extraction from a document.

    This tool searches for the requested keys matching the descriptions in the document.
    """
    logger.debug(f"{document_key=}")
    logger.debug(f"{keys=}")
    logger.debug(f"{descriptions=}")

    if document_key not in local_document_cache:
        doc_keys = ", ".join(local_document_cache.keys())
        raise ValueError(
            f"document-key: {document_key} is not found. Existing document-keys are: {doc_keys}"
        )

    doc = local_document_cache[document_key]
    content_md = doc.export_to_markdown()
    if max_window_size is not None:
        content_md = content_md[:max_window_size]

    prompt = "Extract the following information from the document:\n"
    schema: dict[str, dict[str, str]] = {}
    for key, desc in zip(keys, descriptions, strict=False):
        prompt += f"- {key}: {desc}\n"
        schema[key] = {
            "type": "string",
            "description": desc,
        }
    extraction_schema = {
        "type": "json_schema",
        "json_schema": {
            "name": "extraction",
            "schema": {
                "type": "object",
                "properties": schema,
                "required": list(keys),
            },
        },
    }
    prompt = prompt.strip()

    client = get_llama_stack_client()
    chat_completion = client.chat.completions.create(
        model=settings.extraction_model,
        messages=[
            {
                "role": "system",
                "content": prompt,
            },
            {
                "role": "user",
                "content": content_md,
            },
        ],
        response_format=extraction_schema,
    )  # type: ignore[call-overload, unused-ignore]

    response = json.loads(chat_completion.choices[0].message.content)

    return ExtractedContent(result=response)



"""This module initializes and runs the Docling MCP server."""

import enum
from typing import Annotated

import typer

from docling_mcp.logger import setup_logger
from docling_mcp.shared import mcp

app = typer.Typer()


class ToolGroups(str, enum.Enum):
    """List of available toolsgroups."""

    CONVERSION = "conversion"
    GENERATION = "generation"
    MANIPULATION = "manipulation"
    LLAMA_INDEX_RAG = "llama-index-rag"
    LLAMA_STACK_RAG = "llama-stack-rag"
    LLAMA_STACK_IE = "llama-stack-ie"


class TransportType(str, enum.Enum):
    """List of available protocols."""

    STDIO = "stdio"
    SSE = "sse"
    STREAMABLE_HTTP = "streamable-http"


_DEFAULT_TOOLS = [ToolGroups.CONVERSION, ToolGroups.GENERATION, ToolGroups.MANIPULATION]


@app.command()
def main(
    transport: TransportType = TransportType.STDIO,
    host: str = "localhost",
    port: int = 8000,
    tools: Annotated[
        list[ToolGroups] | None,
        typer.Argument(
            help=f"Tools to be loaded in the server. The default list is {', '.join(_DEFAULT_TOOLS)}"
        ),
    ] = None,
) -> None:
    """Initialize and run the Docling MCP server."""
    # Create a default project logger
    logger = setup_logger()

    if tools is None:
        tools = [*_DEFAULT_TOOLS]

    if ToolGroups.CONVERSION in tools:
        logger.info("loading conversion tools...")
        import docling_mcp.tools.conversion

    if ToolGroups.GENERATION in tools:
        logger.info("loading generation tools...")
        import docling_mcp.tools.generation

    if ToolGroups.MANIPULATION in tools:
        logger.info("loading manipulation tools...")
        import docling_mcp.tools.manipulation

    if ToolGroups.LLAMA_INDEX_RAG in tools:
        logger.info("loading Llama Index RAG tools...")
        import docling_mcp.tools.llama_index.milvus_rag

    if ToolGroups.LLAMA_STACK_RAG in tools:
        logger.info("loading Llama Stack RAG tools...")
        import docling_mcp.tools.llama_stack.rag

    if ToolGroups.LLAMA_STACK_IE in tools:
        logger.info("loading Llama Stack Structured Output tools...")
        import docling_mcp.tools.llama_stack.structured_output

    # Initialize and run the server
    logger.info("starting up Docling MCP-server ...")
    mcp.settings.host = host
    mcp.settings.port = port
    mcp.run(transport=transport.value)


if __name__ == "__main__":
    main()



"""This module contains functions for working with Llama Stack."""

from functools import lru_cache

from llama_stack_client import LlamaStackClient

from docling_mcp.settings.llama_stack import settings


@lru_cache
def get_llama_stack_client() -> LlamaStackClient:
    client = LlamaStackClient(
        base_url=settings.url,
    )
    return client



"""Tools for generating Docling documents."""

import uuid
from dataclasses import dataclass
from io import BytesIO
from typing import Annotated

from mcp.server.fastmcp import Image as MCPImage
from pydantic import Field

from docling.datamodel.base_models import ConversionStatus, InputFormat
from docling.datamodel.document import (
    ConversionResult,
)
from docling.document_converter import DocumentConverter
from docling_core.types.doc.document import (
    ContentLayer,
    DoclingDocument,
    GroupItem,
    LevelNumber,
)
from docling_core.types.doc.labels import (
    DocItemLabel,
    GroupLabel,
)
from docling_core.types.io import DocumentStream

from docling_mcp.docling_cache import get_cache_dir
from docling_mcp.logger import setup_logger
from docling_mcp.shared import local_document_cache, local_stack_cache, mcp

# Create a default project logger
logger = setup_logger()


@dataclass
class NewDoclingDocumentOutput:
    """Output of the create_new_docling_document tool."""

    document_key: Annotated[
        str, Field(description="The unique key that identifies the new document.")
    ]
    prompt: Annotated[str, Field(description="The original prompt.")]


@mcp.tool(title="Create new Docling document")
def create_new_docling_document(
    prompt: Annotated[
        str, Field(description="The prompt text to include in the new document.")
    ],
) -> NewDoclingDocumentOutput:
    """Create a new Docling document from a provided prompt string.

    This function generates a new document in the local document cache with the
    provided prompt text. The document is assigned a unique key derived from an MD5
    hash of the prompt text.
    """
    doc = DoclingDocument(name="Generated Document")

    item = doc.add_text(
        label=DocItemLabel.TEXT,
        text=f"prompt: {prompt}",
        content_layer=ContentLayer.FURNITURE,
    )

    document_key = str(uuid.uuid4()).replace("-", "")

    local_document_cache[document_key] = doc
    local_stack_cache[document_key] = [item]

    return NewDoclingDocumentOutput(document_key, prompt)


@dataclass
class ExportDocumentMarkdownOutput:
    """Output of the export_docling_document_to_markdown tool."""

    document_key: Annotated[
        str,
        Field(description="The unique identifier of the document in the local cache."),
    ]
    markdown: Annotated[
        str, Field(description="The representation of the document in markdown format.")
    ]


@mcp.tool(title="Export Docling document to markdown format")
def export_docling_document_to_markdown(
    document_key: Annotated[
        str,
        Field(description="The unique identifier of the document in the local cache."),
    ],
    max_size: Annotated[
        int | None, Field(description="The maximum number of characters to return.")
    ] = None,
) -> ExportDocumentMarkdownOutput:
    """Export a document from the local document cache to markdown format.

    This tool converts a Docling document that exists in the local cache into
    a markdown formatted string, which can be used for display or further processing.
    """
    if document_key not in local_document_cache:
        doc_keys = ", ".join(local_document_cache.keys())
        raise ValueError(
            f"document-key: {document_key} is not found. Existing document-keys are: {doc_keys}"
        )

    markdown = local_document_cache[document_key].export_to_markdown()
    if max_size:
        markdown = markdown[:max_size]

    return ExportDocumentMarkdownOutput(document_key, markdown)


@dataclass
class SaveDocumentOutput:
    """Output of the save_docling_document tool."""

    md_file: Annotated[
        str,
        Field(
            description="The path in the cache directory to the file in markdown format."
        ),
    ]
    json_file: Annotated[
        str,
        Field(
            description="The path in the cache directory to the file in JSON format."
        ),
    ]


@mcp.tool(title="Save Docling document")
def save_docling_document(
    document_key: Annotated[
        str,
        Field(description="The unique identifier of the document in the local cache."),
    ],
) -> SaveDocumentOutput:
    """Save a document from the local document cache to disk in both markdown and JSON formats.

    This tool takes a document that exists in the local cache and saves it to the specified
    cache directory with filenames based on the document key. Both markdown and JSON versions
    of the document are saved.
    """
    if document_key not in local_document_cache:
        doc_keys = ", ".join(local_document_cache.keys())
        raise ValueError(
            f"document-key: {document_key} is not found. Existing document-keys are: {doc_keys}"
        )

    cache_dir = get_cache_dir()
    md_file = str(cache_dir / f"{document_key}.md")
    json_file = str(cache_dir / f"{document_key}.json")

    local_document_cache[document_key].save_as_markdown(filename=md_file, text_width=72)
    local_document_cache[document_key].save_as_json(filename=json_file)

    return SaveDocumentOutput(md_file, json_file)


@mcp.tool(title="Generate the thumbnail of a page in the Docling document")
def page_thumbnail(
    document_key: Annotated[
        str,
        Field(description="The unique identifier of the document in the local cache."),
    ],
    page_no: Annotated[
        int, Field(description="The number of the page starting at 1")
    ] = 1,
    size: Annotated[
        int, Field(description="The width of the thumbnail in pixels")
    ] = 300,
) -> MCPImage:
    """Generate a thumbnail image for the requested page.

    This tool takes a document that exists in the local cache and generates a thumnail image for the requested page.
    """
    if document_key not in local_document_cache:
        doc_keys = ", ".join(local_document_cache.keys())
        raise ValueError(
            f"document-key: {document_key} is not found. Existing document-keys are: {doc_keys}"
        )

    doc = local_document_cache[document_key]
    if page_no not in doc.pages:
        raise ValueError(
            f"page_no={page_no}: not found in the document. Available pages are: {', '.join(str(k) for k in doc.pages.keys())}"
        )

    im_ref = doc.pages[page_no].image
    if im_ref is None:
        raise ValueError(
            "The DoclingDocument does not have page images. Please configure your server for generating page images using DOCLING_MCP_KEEP_IMAGES=true."
        )
    im = im_ref.pil_image
    if im is None:
        raise RuntimeError("Server error. The image cannot be loaded in PIL.")
    width = size
    scale = float(width) / im.size[0]
    im.thumbnail((width, int(im.size[1] * scale)))

    cache_dir = get_cache_dir()
    im_file = cache_dir / f"{document_key}-{page_no}.png"
    im.save(im_file, format="PNG")

    return MCPImage(path=im_file, format="png")


@dataclass
class UpdateDocumentOutput:
    """Output of the Docling document content generation tools."""

    document_key: Annotated[
        str,
        Field(description="The unique identifier of the document in the local cache."),
    ]


@mcp.tool(title="Add or update title to Docling document")
def add_title_to_docling_document(
    document_key: Annotated[
        str,
        Field(description="The unique identifier of the document in the local cache."),
    ],
    title: Annotated[
        str, Field(description="The title text to add or update to the document.")
    ],
) -> UpdateDocumentOutput:
    """Add or update the title of a document in the local document cache.

    This tool modifies an existing document that has already been processed
    and stored in the local cache. It requires that the document already exists
    in the cache before a title can be added.
    """
    if document_key not in local_document_cache:
        doc_keys = ", ".join(local_document_cache.keys())
        raise ValueError(
            f"document-key: {document_key} is not found. Existing document-keys are: {doc_keys}"
        )

    if len(local_stack_cache[document_key]) == 0:
        raise ValueError(
            f"Stack size is zero for document with document-key: {document_key}. Abort document generation"
        )

    parent = local_stack_cache[document_key][-1]

    if isinstance(parent, GroupItem):
        if parent.label == GroupLabel.LIST or parent.label == GroupLabel.ORDERED_LIST:
            raise ValueError(
                "A list is currently opened. Please close the list before adding a title!"
            )

    item = local_document_cache[document_key].add_title(text=title)
    local_stack_cache[document_key][-1] = item

    return UpdateDocumentOutput(document_key)


@mcp.tool(title="Add section heading to Docling document")
def add_section_heading_to_docling_document(
    document_key: Annotated[
        str,
        Field(description="The unique identifier of the document in the local cache."),
    ],
    section_heading: Annotated[
        str, Field(description="The text to use for the section heading.")
    ],
    section_level: Annotated[
        LevelNumber,
        Field(
            description="The level of the heading, starting from 1, where 1 is the highest level."
        ),
    ],
) -> UpdateDocumentOutput:
    """Add a section heading to an existing document in the local document cache.

    This tool inserts a section heading with the specified heading text and level
    into a document that has already been processed and stored in the local cache.
    Section levels typically represent heading hierarchy (e.g., 1 for H1, 2 for H2).
    """
    if document_key not in local_document_cache:
        doc_keys = ", ".join(local_document_cache.keys())
        raise ValueError(
            f"document-key: {document_key} is not found. Existing document-keys are: {doc_keys}"
        )

    if len(local_stack_cache[document_key]) == 0:
        raise ValueError(
            f"Stack size is zero for document with document-key: {document_key}. Abort document generation"
        )

    parent = local_stack_cache[document_key][-1]

    if isinstance(parent, GroupItem):
        if parent.label == GroupLabel.LIST or parent.label == GroupLabel.ORDERED_LIST:
            raise ValueError(
                "A list is currently opened. Please close the list before adding a section-heading!"
            )

    item = local_document_cache[document_key].add_heading(
        text=section_heading, level=section_level
    )
    local_stack_cache[document_key][-1] = item

    return UpdateDocumentOutput(document_key)


@mcp.tool(title="Add paragraph to Docling document")
def add_paragraph_to_docling_document(
    document_key: Annotated[
        str,
        Field(description="The unique identifier of the document in the local cache."),
    ],
    paragraph: Annotated[
        str, Field(description="The text content to add as a paragraph.")
    ],
) -> UpdateDocumentOutput:
    """Add a paragraph of text to an existing document in the local document cache.

    This tool inserts a new paragraph under the specified section header and level
    into a document that has already been processed and stored in the cache.
    """
    if document_key not in local_document_cache:
        doc_keys = ", ".join(local_document_cache.keys())
        raise ValueError(
            f"document-key: {document_key} is not found. Existing document-keys are: {doc_keys}"
        )

    if len(local_stack_cache[document_key]) == 0:
        raise ValueError(
            f"Stack size is zero for document with document-key: {document_key}. Abort document generation"
        )

    parent = local_stack_cache[document_key][-1]

    if isinstance(parent, GroupItem):
        if parent.label == GroupLabel.LIST or parent.label == GroupLabel.ORDERED_LIST:
            raise ValueError(
                "A list is currently opened. Please close the list before adding a paragraph!"
            )

    item = local_document_cache[document_key].add_text(
        label=DocItemLabel.TEXT, text=paragraph
    )
    local_stack_cache[document_key][-1] = item

    return UpdateDocumentOutput(document_key)


@mcp.tool(title="Open list in Docling document")
def open_list_in_docling_document(
    document_key: Annotated[
        str,
        Field(description="The unique identifier of the document in the local cache."),
    ],
) -> UpdateDocumentOutput:
    """Open a new list group in an existing document in the local document cache.

    This tool creates a new list structure within a document that has already been
    processed and stored in the local cache. It requires that the document already exists
    and that there is at least one item in the document's stack cache.
    """
    if document_key not in local_document_cache:
        doc_keys = ", ".join(local_document_cache.keys())
        raise ValueError(
            f"document-key: {document_key} is not found. Existing document-keys are: {doc_keys}"
        )

    if len(local_stack_cache[document_key]) == 0:
        raise ValueError(
            f"Stack size is zero for document with document-key: {document_key}. Abort document generation"
        )

    item = local_document_cache[document_key].add_group(label=GroupLabel.LIST)
    local_stack_cache[document_key].append(item)

    return UpdateDocumentOutput(document_key)


@mcp.tool(title="Close list in Docling document")
def close_list_in_docling_document(
    document_key: Annotated[
        str,
        Field(description="The unique identifier of the document in the local cache."),
    ],
) -> UpdateDocumentOutput:
    """Closes a list group in an existing document in the local document cache.

    This tool closes a previously opened list structure within a document.
    It requires that the document exists and that there is more than one item
    in the document's stack cache.
    """
    if document_key not in local_document_cache:
        doc_keys = ", ".join(local_document_cache.keys())
        raise ValueError(
            f"document-key: {document_key} is not found. Existing document-keys are: {doc_keys}"
        )

    if len(local_stack_cache[document_key]) <= 1:
        raise ValueError(
            f"Stack size is zero for document with document-key: {document_key}. Abort document generation"
        )

    local_stack_cache[document_key].pop()

    return UpdateDocumentOutput(document_key)


@dataclass
class ListItem:
    """A class to represent a list item pairing."""

    list_item_text: Annotated[str, Field(description="The text of a list item.")]
    list_marker_text: Annotated[str, Field(description="The marker of a list item.")]


@mcp.tool(title="Add items to list in Docling document")
def add_list_items_to_list_in_docling_document(
    document_key: Annotated[
        str,
        Field(description="The unique identifier of the document in the local cache."),
    ],
    list_items: Annotated[
        list[ListItem],
        Field(description="A list of list_item_text and list_marker_text items."),
    ],
) -> UpdateDocumentOutput:
    """Add list items to an open list in an existing document in the local document cache.

    This tool inserts new list items with the specified text and marker into an
    open list within a document. It requires that the document exists and that
    there is at least one item in the document's stack cache.
    """
    if document_key not in local_document_cache:
        doc_keys = ", ".join(local_document_cache.keys())
        raise ValueError(
            f"document-key: {document_key} is not found. Existing document-keys are: {doc_keys}"
        )

    if len(local_stack_cache[document_key]) == 0:
        raise ValueError(
            f"Stack size is zero for document with document-key: {document_key}. Abort document generation"
        )

    parent = local_stack_cache[document_key][-1]

    if isinstance(parent, GroupItem):
        if parent.label != GroupLabel.LIST and parent.label != GroupLabel.ORDERED_LIST:
            raise ValueError(
                "No list is currently opened. Please open a list before adding list-items!"
            )
    else:
        raise ValueError(
            "No list is currently opened. Please open a list before adding list-items!"
        )

    for list_item in list_items:
        local_document_cache[document_key].add_list_item(
            text=list_item.list_item_text,
            marker=list_item.list_marker_text,
            parent=parent,
        )

    return UpdateDocumentOutput(document_key)


@mcp.tool(title="Add HTML table to Docling document")
def add_table_in_html_format_to_docling_document(
    document_key: Annotated[
        str,
        Field(description="The unique identifier of the document in the local cache."),
    ],
    html_table: Annotated[
        str,
        Field(
            description="The HTML string representation of the table to add.",
            examples=[
                "<table><tr><th>Name</th><th>Age</th></tr><tr><td>John</td><td>30</td></tr></table>",
                "<table><tr><th colspan='2'>Demographics</th></tr><tr><th>Name</th><th>Age</th></tr><tr><td>John</td><td rowspan='2'>30</td></tr><tr><td>Jane</td></tr></table>",
            ],
        ),
    ],
    table_captions: Annotated[
        list[str] | None,
        Field(description="A list of caption strings to associate with the table.."),
    ] = None,
    table_footnotes: Annotated[
        list[str] | None,
        Field(description="A list of footnote strings to associate with the table."),
    ] = None,
) -> UpdateDocumentOutput:
    """Add an HTML-formatted table to an existing document in the local document cache.

    This tool parses the provided HTML table string, converts it to a structured table
    representation, and adds it to the specified document. It also supports optional
    captions and footnotes for the table.
    """
    if document_key not in local_document_cache:
        doc_keys = ", ".join(local_document_cache.keys())
        raise ValueError(
            f"document-key: {document_key} is not found. Existing document-keys are: {doc_keys}"
        )

    doc = local_document_cache[document_key]

    if len(local_stack_cache[document_key]) == 0:
        raise ValueError(
            f"Stack size is zero for document with document-key: {document_key}. Abort document generation"
        )

    html_doc: str = f"<html><body>{html_table}</body></html>"

    buff = BytesIO(html_doc.encode("utf-8"))
    doc_stream = DocumentStream(name="tmp", stream=buff)

    converter = DocumentConverter(allowed_formats=[InputFormat.HTML])
    conv_result: ConversionResult = converter.convert(doc_stream)

    if (
        conv_result.status == ConversionStatus.SUCCESS
        and len(conv_result.document.tables) > 0
    ):
        table = doc.add_table(data=conv_result.document.tables[0].data)

        for _ in table_captions or []:
            caption = doc.add_text(label=DocItemLabel.CAPTION, text=_)
            table.captions.append(caption.get_ref())

        for _ in table_footnotes or []:
            footnote = doc.add_text(label=DocItemLabel.FOOTNOTE, text=_)
            table.footnotes.append(footnote.get_ref())
    else:
        raise ValueError(
            "Could not parse the html string of the table! Please fix the html and try again!"
        )

    return UpdateDocumentOutput(document_key)



"""This module defines shared resources."""

from mcp.server.fastmcp import FastMCP

from docling_core.types.doc.document import (
    DoclingDocument,
    NodeItem,
)

# Create a single shared FastMCP instance
mcp = FastMCP("docling")

# Define your shared cache here if it's used by multiple tools
local_document_cache: dict[str, DoclingDocument] = {}
local_stack_cache: dict[str, list[NodeItem]] = {}



"""This module contains the settings for conversion tools."""

from pydantic_settings import BaseSettings, SettingsConfigDict


class Settings(BaseSettings):
    """Settings for the conversion tools."""

    model_config = SettingsConfigDict(
        env_prefix="DOCLING_MCP_",
        env_file=".env",
        # extra="allow",
    )
    keep_images: bool = False


settings = Settings()



"""Tools for manipulating Docling documents."""

import re
from dataclasses import dataclass
from typing import Annotated

from pydantic import Field

from docling_core.types.doc.document import (
    DocItem,
    GroupItem,
    RefItem,
    SectionHeaderItem,
    TextItem,
    TitleItem,
)

from docling_mcp.logger import setup_logger
from docling_mcp.shared import local_document_cache, mcp

# Create a default project logger
logger = setup_logger()


# TODO: Provide a proper structure instead of a single string
@dataclass
class DocumentAnchorOutput:
    """Output of the get_overview_of_document_anchors tool."""

    structure: Annotated[
        str,
        Field(
            description=(
                "A string containing the hierarchical structure of the document with "
                "indentation to show nesting levels, along with anchor references."
            )
        ),
    ]


@mcp.tool(title="Get overview of Docling document anchors")
def get_overview_of_document_anchors(
    document_key: Annotated[
        str,
        Field(description="The unique identifier of the document in the local cache."),
    ],
) -> DocumentAnchorOutput:
    """Retrieve a structured overview of a document from the local document cache.

    This tool returns a text representation of the Docling document's structure,
    showing the hierarchy and types of elements within the document. Each line in the
    output includes the document anchor reference and item label.
    """
    if document_key not in local_document_cache:
        doc_keys = ", ".join(local_document_cache.keys())
        raise ValueError(
            f"document-key: {document_key} is not found. Existing document-keys are: "
            f"{doc_keys}"
        )

    doc = local_document_cache[document_key]

    lines = []
    slevel = 0
    for item, level in doc.iterate_items():
        ref = item.get_ref()

        if isinstance(item, DocItem):
            if isinstance(item, TitleItem):
                lines.append(f"[anchor:{ref.cref}] {item.label}: {item.text}")

            elif isinstance(item, SectionHeaderItem):
                slevel = item.level
                indent = "  " * (level + slevel)
                lines.append(
                    f"{indent}[anchor:{ref.cref}] {item.label}-{level}: {item.text}"
                )

            else:
                indent = "  " * (level + slevel + 1)
                lines.append(f"{indent}[anchor:{ref.cref}] {item.label}")

        elif isinstance(item, GroupItem):
            indent = "  " * (level + slevel + 1)
            lines.append(f"{indent}[anchor:{ref.cref}] {item.label}")

    return DocumentAnchorOutput("\n".join(lines))


@dataclass
class TextSearchOutput:
    """Output of the search_for_text_in_document_anchors tool."""

    result: Annotated[
        str,
        Field(
            description=(
                "A string listing the result of searching for text in the document's "
                "anchors. If matches were found, the result indicates what text matched "
                "at which anchors, along with the number of occurrences. If no matches "
                "were found, the result indicates that no matches were found."
            )
        ),
    ]


@mcp.tool(title="Search for text in Docling document anchors")
def search_for_text_in_document_anchors(
    document_key: Annotated[
        str,
        Field(description="The unique identifier of the document in the local cache."),
    ],
    text: Annotated[
        str,
        Field(
            description="The string of text to search for in the document's anchors."
        ),
    ],
) -> TextSearchOutput:
    """Search for specific text and keywords within a document's anchors.

    This tool takes a string of text to search for and returns a string of all
    document anchors that contain the exact text. The search is case-insensitive.
    If the exact text is not found, the tool will search for individual keywords
    within the text, splitting it on non-alphanumeric characters. If keywords
    are found, they are listed alongside their number of occurrences in parentheses.
    """
    if document_key not in local_document_cache:
        doc_keys = ", ".join(local_document_cache.keys())
        raise ValueError(
            f"document-key: {document_key} is not found. Existing document-keys are: {doc_keys}"
        )

    doc = local_document_cache[document_key]
    exact_matches = []
    matches = []
    keywords_set = {word for word in set(re.findall(r"\b\w+\b", text.lower())) if word}

    for item, _ in doc.iterate_items():
        if isinstance(item, TextItem):
            ref = item.get_ref()

            if text.lower() in item.text.lower():
                exact_matches.append(f"[anchor:{ref.cref}]")

            if not exact_matches:
                keyword_occurrences: dict[str, int] = {}
                total_matches = 0

                strings = re.findall(r"\b\w+\b", item.text.lower())

                for string in strings:
                    if string in keywords_set:
                        total_matches += 1
                        if string in keyword_occurrences:
                            keyword_occurrences[string] += 1
                        else:
                            keyword_occurrences[string] = 1

                if keyword_occurrences:
                    matches.append(
                        (
                            f"[anchor:{ref.cref}] keyword matches ({total_matches} total):{','.join([f' {k} ({v} occurrences)' for k, v in sorted(keyword_occurrences.items(), key=lambda x: x[1], reverse=True)])}",
                            total_matches,
                        )
                    )

    if exact_matches:
        return TextSearchOutput(
            "Found exact text matches in the following anchors:\n"
            + "\n".join(exact_matches)
        )
    if matches:
        return TextSearchOutput(
            "No exact text matches were found. Found individual keyword matches in the following anchors:\n"
            + "\n".join(
                [
                    match[0]
                    for match in sorted(
                        matches, key=lambda match: match[1], reverse=True
                    )
                ]
            )
        )
    return TextSearchOutput(
        f"No exact text matches nor individual keyword matches found for '{text}' in document with key {document_key}."
    )


@dataclass
class DocumentItemText:
    """Text content of a Docling document item."""

    text: Annotated[
        str,
        Field(description="The text content of an item in a Docling document."),
    ]


@mcp.tool(title="Get text of Docling document item at anchor")
def get_text_of_document_item_at_anchor(
    document_key: Annotated[
        str,
        Field(description="The unique identifier of the document in the local cache."),
    ],
    document_anchor: Annotated[
        str,
        Field(
            description=(
                "The anchor reference that identifies the specific item within the "
                "document."
            ),
            examples=["#/texts/2"],
        ),
    ],
) -> DocumentItemText:
    """Retrieve the text content of a specific document item identified by its anchor.

    This tool extracts the text from a Docling document item at the specified anchor
    location within a document that exists in the local document cache.
    """
    if document_key not in local_document_cache:
        doc_keys = ", ".join(local_document_cache.keys())
        raise ValueError(
            f"document-key: {document_key} is not found. Existing document-keys are: {doc_keys}"
        )

    doc = local_document_cache[document_key]

    ref = RefItem(cref=document_anchor)
    item = ref.resolve(doc=doc)

    if isinstance(item, TextItem):
        text = item.text
    else:
        raise ValueError(
            f"Item at {document_anchor} for document-key: {document_key} is not a textual item."
        )

    return DocumentItemText(text)


@dataclass
class UpdateDocumentOutput:
    """Output of the Docling document content modification tools."""

    document_key: Annotated[
        str,
        Field(description="The unique identifier of the document in the local cache."),
    ]


@mcp.tool(title="Update text of Docling document item at anchor")
def update_text_of_document_item_at_anchor(
    document_key: Annotated[
        str,
        Field(description="The unique identifier of the document in the local cache."),
    ],
    document_anchor: Annotated[
        str,
        Field(
            description=(
                "The anchor reference that identifies the specific item within the "
                "document."
            ),
            examples=["#/texts/6"],
        ),
    ],
    updated_text: Annotated[
        str,
        Field(description="The new text content to replace the existing content."),
    ],
) -> UpdateDocumentOutput:
    """Update the text content of a specific document item identified by its anchor.

    This tool modifies the text of an existing document item at the specified anchor
    location within a document that exists in the local document cache. It requires
    that the document already exists in the cache before a modification can be made.
    """
    if document_key not in local_document_cache:
        doc_keys = ", ".join(local_document_cache.keys())
        raise ValueError(
            f"document-key: {document_key} is not found. Existing document-keys are: "
            f"{doc_keys}"
        )

    doc = local_document_cache[document_key]

    ref = RefItem(cref=document_anchor)
    item = ref.resolve(doc=doc)

    if isinstance(item, TextItem):
        item.text = updated_text
    else:
        raise ValueError(
            f"Item at {document_anchor} for document-key: {document_key} is not a "
            "textual item."
        )

    return UpdateDocumentOutput(document_key=document_key)


@mcp.tool(title="Delete Docling document items at anchors")
def delete_document_items_at_anchors(
    document_key: Annotated[
        str,
        Field(description="The unique identifier of the document in the local cache."),
    ],
    document_anchors: Annotated[
        list[str],
        Field(
            description=(
                "A list of anchor references identifying the items to be deleted from the "
                "document."
            ),
            examples=["#/texts/2", "#/tables/1"],
        ),
    ],
) -> UpdateDocumentOutput:
    """Delete multiple document items identified by their anchors.

    This tool removes specified items from a Docling document that exists in the local
    document cache, based on their anchor references. It requires that the document
    already exists in the cache before performing the deletion.
    """
    if document_key not in local_document_cache:
        doc_keys = ", ".join(local_document_cache.keys())
        raise ValueError(
            f"document-key: {document_key} is not found. Existing document-keys are: {doc_keys}"
        )

    doc = local_document_cache[document_key]

    items = []
    for _ in document_anchors:
        ref = RefItem(cref=_)
        items.append(ref.resolve(doc=doc))

    doc.delete_items(node_items=items)

    return UpdateDocumentOutput(document_key=document_key)



"""This module contains the settings for the Llama Index usages."""

from pydantic_settings import BaseSettings, SettingsConfigDict


class Settings(BaseSettings):
    """Settings for the Llama Index usages."""

    model_config = SettingsConfigDict(
        env_prefix="DOCLING_MCP_LI_",
        env_file=".env",
        # extra="allow",
    )
    api_base: str = "http://127.0.0.1:1234/v1"
    api_key: str = "none"
    model_id: str = "ibm/granite-3.2-8b"
    embedding_model: str = "BAAI/bge-base-en-v1.5"


settings = Settings()



"""This module defines the shared components for the Llama Index tools tools."""

from llama_index.core import Settings as LISettings
from llama_index.core.indices.vector_store.base import VectorStoreIndex
from llama_index.embeddings.huggingface import HuggingFaceEmbedding
from llama_index.llms.openai_like import OpenAILike
from llama_index.node_parser.docling import DoclingNodeParser
from llama_index.vector_stores.milvus import MilvusVectorStore

from docling_mcp.settings.llama_index import settings

embed_model = HuggingFaceEmbedding(model_name=settings.embedding_model)
LISettings.embed_model = embed_model
LISettings.llm = OpenAILike(
    api_base=settings.api_base,
    api_key=settings.api_key,
    model=settings.model_id,
    request_timeout=120.0,
)

node_parser = DoclingNodeParser()

embed_dim = len(embed_model.get_text_embedding("hi"))

milvus_vector_store = MilvusVectorStore(
    uri="./milvus_demo.db", dim=embed_dim, overwrite=True
)

local_index_cache: dict[str, VectorStoreIndex] = {}



"""This module defines the RAG with milvus tools."""

import json
from dataclasses import dataclass
from typing import Annotated, Any

from llama_index.core import Document, StorageContext, VectorStoreIndex
from llama_index.core.base.response.schema import (
    RESPONSE_TYPE,
    Response,
)
from mcp.shared.exceptions import McpError
from mcp.types import INTERNAL_ERROR, ErrorData
from pydantic import Field

from docling_core.types.doc.document import DoclingDocument

from docling_mcp.logger import setup_logger
from docling_mcp.shared import (
    local_document_cache,
    mcp,
)
from docling_mcp.tools.llama_index._shared import (
    local_index_cache,
    milvus_vector_store,
    node_parser,
)

logger = setup_logger()


@mcp.tool(title="Export Docling document to vector database")
def export_docling_document_to_vector_db(
    document_key: Annotated[
        str,
        Field(description="The unique identifier of the document in the local cache."),
    ],
) -> str:
    """Exports a document from the local document cache to a vector database for search capabilities.

    This tool converts a Docling document that exists in the local cache into markdown format,
    then loads it into a vector database index. This allows the document to be searched using
    semantic search techniques.

    Raises:
        ValueError: If the specified document_key does not exist in the local cache.
    """
    if document_key not in local_document_cache:
        doc_keys = ", ".join(local_document_cache.keys())
        raise ValueError(
            f"document-key: {document_key} is not found. Existing document-keys are: {doc_keys}"
        )

    docling_document: DoclingDocument = local_document_cache[document_key]
    document_dict: dict[str, Any] = docling_document.export_to_dict()
    document_json: str = json.dumps(document_dict)

    document = Document(
        text=document_json,
        metadata={"filename": docling_document.name},
    )

    index = VectorStoreIndex.from_documents(
        documents=[document],
        transformations=[node_parser],
        storage_context=StorageContext.from_defaults(vector_store=milvus_vector_store),
    )

    index.insert(document)

    local_index_cache["milvus_index"] = index

    return f"Successful initialisation for document with id {document_key}"


@dataclass
class SearchDocumentOutput:
    """Output of the search documents tool."""

    answer: Annotated[
        str,
        Field(
            description="A string containing the relevant contextual information retrieved from the documents that best matches the query."
        ),
    ]

    # TODO: future updates could provide the grounding elements metadata


@mcp.tool(title="Search query in documents")
def search_documents(
    query: Annotated[
        str,
        Field(
            description="The search query text used to find relevant information in the indexed documents."
        ),
    ],
) -> SearchDocumentOutput:
    """Searches through previously uploaded and indexed documents using semantic search.

    This function retrieves relevant information from documents that have been processed
    and added to the vector database. It uses semantic similarity to find content that
    best matches the query, rather than simple keyword matching.
    """
    index = local_index_cache["milvus_index"]

    query_engine = index.as_query_engine()
    response: RESPONSE_TYPE = query_engine.query(query)

    if isinstance(response, Response):
        if response.response is not None:
            return SearchDocumentOutput(answer=response.response)
        else:
            raise McpError(
                ErrorData(
                    code=INTERNAL_ERROR,
                    message="Response object has no response content",
                )
            )
    else:
        raise McpError(
            ErrorData(
                code=INTERNAL_ERROR,
                message=f"Unexpected response type: {type(response)}",
            )
        )



"""Utility module for logging."""

import logging


def setup_logger() -> logging.Logger:
    """Setup and return a logger for the entire project."""
    # Create logger
    logger = logging.getLogger("docling_mcp")
    logger.setLevel(logging.INFO)

    # Create a handler and set its level to INFO
    handler = logging.StreamHandler()
    handler.setLevel(logging.INFO)

    # Create a formatter and add it to the handler
    formatter = logging.Formatter(
        "%(asctime)s - %(name)s - %(levelname)s - %(message)s"
    )
    handler.setFormatter(formatter)

    # Add the handler to the logger
    logger.addHandler(handler)

    return logger



"""Tools for converting documents into DoclingDocument objects."""

import gc
from dataclasses import dataclass
from functools import lru_cache
from pathlib import Path
from typing import Annotated

from mcp.server.fastmcp import Context
from mcp.shared.exceptions import McpError
from mcp.types import INTERNAL_ERROR, ErrorData
from pydantic import Field

from docling.datamodel.base_models import InputFormat
from docling.datamodel.pipeline_options import (
    PdfPipelineOptions,
)
from docling.document_converter import DocumentConverter, FormatOption, PdfFormatOption
from docling_core.types.doc.document import (
    ContentLayer,
)
from docling_core.types.doc.labels import (
    DocItemLabel,
)

from docling_mcp.docling_cache import get_cache_key
from docling_mcp.logger import setup_logger
from docling_mcp.settings.conversion import settings
from docling_mcp.shared import local_document_cache, local_stack_cache, mcp

# Create a default project logger
logger = setup_logger()


def cleanup_memory() -> None:
    """Force garbage collection to free up memory."""
    logger.info("Performed memory cleanup")
    gc.collect()


@dataclass
class IsDoclingDocumentInCacheOutput:
    """Output of the is_document_in_local_cache tool."""

    in_cache: Annotated[
        bool,
        Field(
            description=(
                "Whether the document is already converted and in the local cache."
            )
        ),
    ]


@mcp.tool(title="Is Docling document in cache")
def is_document_in_local_cache(
    document_key: Annotated[
        str,
        Field(description="The unique identifier of the document in the local cache."),
    ],
) -> IsDoclingDocumentInCacheOutput:
    """Verify if a Docling document is already converted and in the local cache."""
    return IsDoclingDocumentInCacheOutput(document_key in local_document_cache)


@dataclass
class ConvertDocumentOutput:
    """Output of the convert_document_into_docling_document tool."""

    from_cache: Annotated[
        bool,
        Field(
            description=(
                "Whether the document was already converted in the local cache."
            )
        ),
    ]
    document_key: Annotated[
        str,
        Field(description="The unique identifier of the document in the local cache."),
    ]


@lru_cache
def _get_converter() -> DocumentConverter:
    pipeline_options = PdfPipelineOptions()
    # pipeline_options.do_ocr = False  # Skip OCR for faster processing (enable for scanned docs)
    pipeline_options.generate_page_images = settings.keep_images

    format_options: dict[InputFormat, FormatOption] = {
        InputFormat.PDF: PdfFormatOption(pipeline_options=pipeline_options),
        InputFormat.IMAGE: PdfFormatOption(pipeline_options=pipeline_options),
    }

    logger.info(f"Creating DocumentConverter with format_options: {format_options}")
    return DocumentConverter(format_options=format_options)


@mcp.tool(title="Convert document into Docling document")
def convert_document_into_docling_document(
    source: Annotated[
        str,
        Field(description="The URL or local file path to the document."),
    ],
) -> ConvertDocumentOutput:
    """Convert a document of any type from a URL or local path and store in local cache.

    This tool takes a document's URL or local file path, converts it using
    Docling's DocumentConverter, and stores the resulting Docling document in a
    local cache. It returns an output with a boolean set to False along with the
    document's unique cache key. If the document was already in the local cache,
    the conversion is skipped and the output boolean is set to True.
    """
    try:
        # Remove any quotes from the source string
        source = source.strip("\"'")

        # Log the cleaned source
        logger.info(f"Processing document from source: {source}")

        # Generate cache key
        cache_key = get_cache_key(source)

        if cache_key in local_document_cache:
            logger.info(f"{source} has previously been added.")
            return ConvertDocumentOutput(True, cache_key)

        # Get converter
        converter = _get_converter()

        # Convert the document
        logger.info("Start conversion")
        result = converter.convert(source)

        # Check for errors - handle different API versions
        has_error = False
        error_message = ""

        # Try different ways to check for errors based on the API version
        if hasattr(result, "status"):
            if hasattr(result.status, "is_error"):
                has_error = result.status.is_error
            elif hasattr(result.status, "error"):
                has_error = result.status.error

        if hasattr(result, "errors") and result.errors:
            has_error = True
            error_message = str(result.errors)

        if has_error:
            error_msg = f"Conversion failed: {error_message}"
            raise McpError(ErrorData(code=INTERNAL_ERROR, message=error_msg))

        local_document_cache[cache_key] = result.document

        item = result.document.add_text(
            label=DocItemLabel.TEXT,
            text=f"source: {source}",
            content_layer=ContentLayer.FURNITURE,
        )

        local_stack_cache[cache_key] = [item]

        # Log completion
        logger.info(f"Successfully created the Docling document: {source}")

        # Clean up memory
        cleanup_memory()

        return ConvertDocumentOutput(False, cache_key)

    except Exception as e:
        logger.exception(f"Error converting document: {source}")
        raise McpError(
            ErrorData(code=INTERNAL_ERROR, message=f"Unexpected error: {e!s}")
        ) from e


@mcp.tool(
    title="Convert files from directory into Docling document", structured_output=True
)
async def convert_directory_files_into_docling_document(
    source: Annotated[
        str,
        Field(description="The path to a local directory"),
    ],
    ctx: Context,  # type: ignore[type-arg]
) -> list[ConvertDocumentOutput]:
    """Convert all files from a local directory path and store them in local cache.

    This tool takes a local directory path, converts every file in the directory using
    Docling's DocumentConverter and stores the resulting Docling documents in a local
    cache. It returns a list of conversion outputs, where each output consists of a
    boolean set to False along with a document's unique cache key. If a document was
    already in the local cache, the conversion is skipped and the output boolean is set
    to True.
    """
    try:
        # Remove any quotes from the source string
        source = source.strip("\"'")
        directory = Path(source)
        files: list[Path] = list(directory.iterdir())
        out: list[ConvertDocumentOutput] = []
        logger.info("Getting the converter")
        converter = _get_converter()

        for i, file in enumerate(files):
            if not file.is_file():
                continue

            # Track progress
            await ctx.info(f"Processing file {file}")
            await ctx.report_progress(i + 1, len(files))

            logger.info(f"Processing file {file}")
            cache_key = get_cache_key(str(file))
            if cache_key in local_document_cache:
                logger.info(f"{file} has been previously converted.")
                out.append(ConvertDocumentOutput(True, cache_key))
            else:
                # Convert the document
                logger.info("Start conversion")
                result = converter.convert(file)
                has_error = False
                error_message = ""
                if hasattr(result, "status"):
                    if hasattr(result.status, "is_error"):
                        has_error = result.status.is_error
                    elif hasattr(result.status, "error"):
                        has_error = result.status.error

                if hasattr(result, "errors") and result.errors:
                    has_error = True
                    error_message = str(result.errors)

                if has_error:
                    error_msg = f"Conversion failed: {error_message}"
                    raise McpError(ErrorData(code=INTERNAL_ERROR, message=error_msg))

                local_document_cache[cache_key] = result.document

                item = result.document.add_text(
                    label=DocItemLabel.TEXT,
                    text=f"source: {file}",
                    content_layer=ContentLayer.FURNITURE,
                )

                local_stack_cache[cache_key] = [item]

                await ctx.debug(
                    f"Completed step {i + 1} with Docling document key: {cache_key}"
                )
                logger.info(f"Successfully created the Docling document: {file}")
                out.append(ConvertDocumentOutput(False, cache_key))

        cleanup_memory()

        return out

    except Exception as e:
        logger.exception(f"Error converting files in directory: {source}")
        raise McpError(
            ErrorData(code=INTERNAL_ERROR, message=f"Unexpected error: {e!s}")
        ) from e
